import re
import uuid
import json
import traceback
from pathlib import Path
from urllib.parse import urlparse
from lyrebird.log import get_logger
from jinja2 import Template
from lyrebird.application import config
from lyrebird.mock.handlers import snapshot_helper
from lyrebird.mock.dm.jsonpath import jsonpath


PROP_FILE_NAME = '.lyrebird_prop'
logger = get_logger()


class DataManager:

    def __init__(self):
        self.root_path: Path = None
        self.display_data_map = {}
        self.id_map = {}
        self.activated_data = {}
        self.activated_group = {}
        self.secondary_activated_data = {}
        self.LEVEL_SECONDARY_ACTIVATED = 2
        self.clipboard = None
        self.save_to_group_id = None
        self.tmp_group = {'id': 'tmp_group', 'type': 'group', 'name': 'tmp-group', 'label': [], 'children': []}
        self.snapshot_helper = snapshot_helper.SnapshotHelper()
        self.root = self.get_default_root()

    def get_default_root(self):
        return {
            'id': str(uuid.uuid4()),
            'name': '$',
            'type': 'group',
            'parent_id': None,
            'label': [],
            'children': []
        }

    def set_adapter(self, adapter_cls):
        self._adapter = adapter_cls(self)

    def set_root(self, uri):
        self._adapter._set_root(uri)

    def reload(self):
        self._adapter._reload()

    def get(self, _id):
        """
        Get mock group or data by id
        """
        if not self.root:
            raise RootNotSet
        node = self.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)
        if node.get('type') == 'group' or node.get('type') == None:
            return node
        elif node.get('type') == 'data':
            return self._adapter._load_data(_id)

    # -----
    # Mock operations
    # -----

    def make_data_map_by_group(self, group_ids):
        self.display_data_map = {k:v for k,v in self.root.items() if k!='children'}
        self.display_data_map['children'] = []

        for group_id in group_ids:
            map_pointer = self.display_data_map['children']

            group = self.id_map.get(group_id)
            group_abs_parent_obj = self._get_abs_parent_obj(group)

            # init parent
            # the last node in group['abs_parent_path'] is node itself, expand node itself
            # expand root which is already in the data_map
            for group_info in group_abs_parent_obj[1:-1]:
                parent_id = group_info['id']

                for index, child in enumerate(map_pointer):
                    if parent_id == child['id']:
                        map_pointer = map_pointer[index]['children']
                        break

                else:
                    node = self.id_map.get(parent_id)

                    new_node = {k:v for k,v in node.items() if k != 'children'}
                    new_node['children'] = []
                    map_pointer.append(new_node)
                    # Because of new_node is `append` into map_pointer, new_node is the last child of map_pointer
                    # move map_pointer first, sort after
                    map_pointer = map_pointer[-1]['children']
                    map_pointer.sort(key=lambda x:x.get('name', ''))

            # update node and node's children
            node = self.id_map.get(group_id)
            map_pointer.append(node)

        return self.display_data_map

    def activate(self, _id):
        """
        Activite data by id
        """
        if not self.root:
            raise RootNotSet
        _node = self.id_map.get(_id)
        if _node:
            self._activate(_node)
        else:
            raise IDNotFound(f'ID:{_id}')
        self._activate_super_node(_node, level_lefted=self.LEVEL_SECONDARY_ACTIVATED)
        self.activated_group[_id] = _node

    def _activate_super_node(self, node, level_lefted=1):
        if level_lefted <= 0:
            return
        if not node.get('super_id'):
            return
        _secondary_search_node_id = node.get('super_id')
        _secondary_search_node = self.id_map.get(_secondary_search_node_id)
        if not _secondary_search_node:
            raise IDNotFound(f'Secondary search node ID: {_secondary_search_node_id}')
        self._activate(_secondary_search_node, secondary_search=True)
        self._activate_super_node(_secondary_search_node, level_lefted=level_lefted-1)

    def _activate(self, node, secondary_search=False):
        if node.get('type', '') == 'data':
            _mock_data = self._adapter._load_data(node['id'])
            if _mock_data:
                if not secondary_search:
                    self.activated_data[node['id']] = _mock_data
                else:
                    self.secondary_activated_data[node['id']] = _mock_data
        elif node.get('type', '') == 'group':
            if 'children' in node:
                for child in node['children']:
                    self._activate(child, secondary_search=secondary_search)

    def deactivate(self):
        """
        Clear activated data
        """
        self.activated_data = {}
        self.activated_group = {}
        self.secondary_activated_data = {}

    def reactive(self):
        origin_activated_group = self.activated_group
        self.deactivate()
        for _group_id in origin_activated_group:
            self.activate(_group_id)

    def get_matched_data(self, flow):
        """
        Find matched mock data from activated data
        """
        _matched_data = []
        for _data_id in self.activated_data:
            _data = self.activated_data[_data_id]
            if self._is_match_rule(flow, _data.get('rule')):
                _matched_data.append(_data)
        if len(_matched_data) <= 0:
            for _data_id in self.secondary_activated_data:
                _data = self.secondary_activated_data[_data_id]
                if self._is_match_rule(flow, _data.get('rule')):
                    _matched_data.append(_data)

        # TODO render mock data before response, support more functions
        params = {
            'ip': config.get('ip'),
            'port': config.get('mock.port')
        }
        for response_data in _matched_data:
            if 'response' not in response_data:
                continue
            if 'data' not in response_data['response']:
                continue
            if not response_data['response']['data']:
                continue
            resp_data_template = Template(response_data['response']['data'])
            response_data['response']['data'] = resp_data_template.render(params)

        return _matched_data

    def _is_match_rule(self, flow, rules):
        if not rules:
            return False
        for rule_key, pattern in rules.items():
            targets = self._get_rule_targets(rule_key, flow)
            if not targets:
                return False
            if not self._is_target_pattern_matched(pattern, targets):
                return False
        return True

    def _get_rule_targets(self, rule_key, flow):
        search_res = jsonpath.search(flow, rule_key)
        if not search_res:
            return None
        return [s.node for s in search_res if s.node]

    def _is_target_pattern_matched(self, pattern, targets):
        for target in targets:
            try:
                search_result = re.search(pattern, target)
            except:
                logger.warning(f'Illegal regular match in mock data!\n {traceback.format_exc()}')
                return False
            if not search_result:
                return False
        return True

    # -----
    # Data tree operations
    # -----

    def _get_request_path(self, request):
        path = request.get('path')
        if not path:
            if not request.get('url'):
                return ''
            parsed_url = urlparse(request['url'])
            path = parsed_url.path
        return path

    def add_data(self, parent_id, raw_data, **kwargs):
        if not isinstance(raw_data, dict):
            raise DataObjectSouldBeADict

        parent_node = None
        if parent_id == 'tmp_group':
            parent_node = self.tmp_group
        elif parent_id:
            parent_node = self.id_map.get(parent_id)
            if not parent_node:
                raise IDNotFound(parent_id)
            if parent_node['type'] == 'data':
                raise DataObjectCannotContainAnyOtherObject

        data = dict(raw_data)
        _data_id = kwargs.get('data_id') or str(uuid.uuid4())
        data['id'] = _data_id
        if 'request' in data:
            # TODO remove it with inspector frontend
            data['request'] = dict(raw_data['request'])

            _data_name = self._get_request_path(data['request'])
            _data_rule = {
                'request.url': f'(?=.*{self._get_request_path(data["request"])})'
            }
            if 'data' in data['request']:
                data['request']['data'] = self._flow_data_2_str(data['request']['data'])
        else:
            _data_name = data.get('name')
            _data_rule = {'request.url': '(?=.*YOUR-REQUEST-PATH)(?=.*PARAMS)'}
            data['request'] = {}

        if 'response' in data:
            # TODO remove it with inspector frontend
            data['response'] = dict(raw_data['response'])

            if 'data' in data['response']:
                data['response']['data'] = self._flow_data_2_str(data['response']['data'])
        else:
            data['response'] = {}

        # proxy_response will not be saved
        if 'proxy_response' in data:
            del data['proxy_response']

        data['name'] = _data_name
        data['rule'] = _data_rule

        data['parent_id'] = parent_id

        output_path = (kwargs['output'] / _data_id) if kwargs.get('output') else None

        self._adapter._add_data(data, path=output_path)

        if parent_node:
            data_node = {}
            data_node['id'] = _data_id
            data_node['name'] = _data_name
            data_node['type'] = 'data'
            data_node['parent_id'] = parent_id

            # New data added in the head of child list
            parent_node['children'].insert(0, data_node)
            logger.debug(f'*** Add to node {data_node}')
            # Update ID mapping
            self.id_map[_data_id] = data_node
            self._adapter._add_group(data_node)

        return _data_id

    def _flow_data_2_str(self, data):
        if isinstance(data, str):
            return data
        return json.dumps(data, ensure_ascii=False)

    def add_group(self, parent_id, name):
        if parent_id == None:
            parent_node = self.root
        else:
            parent_node = self.id_map.get(parent_id)
        if not parent_node:
            raise IDNotFound(parent_id)
        if parent_node.get('type') == 'data':
            raise DataObjectCannotContainAnyOtherObject
        # Add group
        group_id = str(uuid.uuid4())
        new_group = {
            'id': group_id,
            'name': name,
            'type': 'group',
            'parent_id': parent_id,
            'label': [],
            'children': [],
            'super_id': None
        }
        # New group added in the head of child list
        if 'children' not in parent_node:
            parent_node['children'] = []
        parent_node['children'].insert(0, new_group)
        # Register ID
        self.id_map[group_id] = new_group
        # Save prop
        self._adapter._add_group(new_group)
        return group_id

    def add_group_by_path(self, path):
        parent_name = path.strip('/')
        current = self.root
        if not parent_name:
            return current['id']
        parent_name_list = parent_name.split('/')
        for name in parent_name_list:
            for child in current['children']:
                if name == child['name']:
                    current = child
                    break
            else:
                group_id = self.add_group(current['id'], name)
                current = self.id_map.get(group_id)
        return current['id']

    def delete(self, _id):
        target_node = self.id_map.get(_id)
        if not target_node:
            raise IDNotFound(_id)
        parent_id = target_node.get('parent_id')
        # Remove refs
        if parent_id:
            parent = self.id_map.get(parent_id)
            parent['children'].remove(target_node)
        else:
            self.root['children'].remove(target_node)
        self._delete(_id)

    def _delete(self, _id):
        target_node = self.id_map.get(_id)
        if not target_node:
            raise IDNotFound(_id)
        if 'children' in target_node and len(target_node['children']) > 0:
            for child in target_node['children'][::-1]:
                self._delete(child['id'])
                target_node['children'].pop(-1)
        # Remove from activated_group
        if _id in self.activated_group:
            self.activated_group.pop(_id)
        # Delete from ID mapping
        self.id_map.pop(_id)
        # Delete from mock tree
        self._adapter._delete_group(_id)
        # Delete from file system
        if target_node['type'] == 'data':
            self._adapter._delete_data(_id)

    def cut(self, _id):
        _node = self.id_map.get(_id)
        if not _node:
            raise IDNotFound(_id)
        self.clipboard = {
            'type': 'cut',
            'id': _id,
            'node': _node
        }

    def copy(self, _id):
        _node = self.id_map.get(_id)
        if not _node:
            raise IDNotFound(_id)
        self.clipboard = {
            'type': 'copy',
            'id': _id,
            'node': _node
        }

    def import_(self, node):
        self.clipboard = {
            'type': 'import',
            'id': node['id'],
            'node': node
        }

    def paste(self, parent_id, **kwargs):
        if not self.clipboard:
            raise NoneClipbord
        _parent_node = self.id_map.get(parent_id)
        _node = self.clipboard['node']
        if not _parent_node:
            raise IDNotFound(parent_id)
        if not _parent_node.get('children'):
            _parent_node['children'] = []
        if self.clipboard['type'] == 'cut':
            _origin_parent = self.id_map.get(_node['parent_id'])
            _origin_parent['children'].remove(_node)
            _parent_node['children'].insert(0, _node)
            _node['parent_id'] = parent_id
            _group_id = _node['id']
            self._adapter._update_group(_node)
        elif self.clipboard['type'] == 'copy':
            new_name = self._get_copy_node_new_name(_node)
            _group_id = self._copy_node(_parent_node, _node, name=new_name, origin_name=_node['name'], **kwargs)
        elif self.clipboard['type'] == 'import':
            _group_id = self._copy_node(_parent_node, _node, **kwargs)
        return _group_id

    def _copy_node(self, parent_node, node, **kwargs):
        new_node = {}
        new_node.update(node)
        new_node['id'] = str(uuid.uuid4())
        new_node['parent_id'] = parent_node['id']
        new_node['name'] = kwargs['name'] if kwargs.get('name') else new_node['name']
        # Add to target node
        if not parent_node.get('children'):
            parent_node['children'] = []
        parent_node['children'].insert(0, new_node)
        # Add node
        self._adapter._add_group(new_node, **kwargs)
        # Register ID
        self.id_map[new_node['id']] = new_node
        if new_node['type'] == 'group':
            kwargs.pop('name') if kwargs.get('name') else None
            kwargs.pop('origin_name') if kwargs.get('origin_name') else None
            new_node['children'] = []
            for child in node['children']:
                self._copy_node(new_node, child, **kwargs)
        elif new_node['type'] == 'data':
            self._copy_file(new_node, node, **kwargs)
        return new_node['id']

    def _copy_file(self, target_data_node, data_node, **kwargs):
        _id = data_node['id']
        custom_root = kwargs.get('custom_input_file_path')
        if not custom_root:
            prop = self._adapter._load_data(_id)
        else:
            prop = self._adapter._load_data(None, path=(Path(custom_root)/_id))
        new_file_id = target_data_node['id']
        prop['id'] = new_file_id
        prop['name'] = kwargs.pop('name') if kwargs.get('name') else prop['name']
        self._adapter._add_data(prop)

    def _get_copy_node_new_name(self, _node):
        COPY_NODE_NAME_SUFFIX = ' - copy'
        return _node['name'] + COPY_NODE_NAME_SUFFIX

    def _sort_children_by_name(self):
        for node_id in self.id_map:
            node = self.id_map[node_id]
            if 'children' not in node:
                # fix mock data group has no children
                if node['type'] == 'group':
                    node['children'] = []
                continue
            node['children'] = sorted(node['children'], key=lambda sub_node: sub_node['name'])

    # -----
    # Conflict checker
    # -----

    def check_conflict(self, _id):
        node = self.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)
        data_array = []

        def _read_data(node):
            if node.get('type') == 'data':
                _data = self._adapter._load_data(node['id'])
                _data['parent_id'] = node['parent_id']
                data_array.append(_data)
            elif node.get('type') == 'group':
                for child in node['children']:
                    _read_data(child)
        _read_data(node)
        return self.check_conflict_data(data_array)

    def activated_data_check_conflict(self):
        data_array = list(self.activated_data.values())
        return self.check_conflict_data(data_array)

    def check_conflict_data(self, data_array):
        conflict_rules = []
        for _data in data_array:
            _rule = _data['rule']
            _hit_data = []
            for _test_data in data_array:
                if self._is_match_rule(_test_data, _rule):
                    _target_node = {
                        'id': _test_data['id'],
                        'name': _test_data['name'],
                        'url': _test_data['request']['url'],
                        'abs_parent_path': self._get_abs_parent_path(_test_data)
                    }
                    _hit_data.append(_target_node)
            if len(_hit_data) > 1:
                _src_node = {
                    'id': _data['id'],
                    'name': _data['name'],
                    'rule': _data['rule'],
                    'abs_parent_path': self._get_abs_parent_path(_data)
                }
                conflict_rules.append(
                    {
                        'data': _src_node,
                        'conflict_data': _hit_data
                    }
                )
        return conflict_rules

    def _get_abs_parent_path(self, node, path='/'):
        parent_node = self._get_node_parent(node)
        if parent_node is None:
            return path
        current_path = '/' + node['name'] + path
        return self._get_abs_parent_path(parent_node, path=current_path)

    def _get_abs_parent_obj(self, node, parent_obj=None):
        if parent_obj is None:
            parent_obj = []
        if 'id' not in node:
            return parent_obj
        node_info = self.id_map.get(node['id'])
        if not node_info:
            return parent_obj
        parent_obj.insert(0, {
            'id': node_info['id'],
            'name': node_info['name'],
            'type': node_info['type'],
            'parent_id': node_info['parent_id']
        })
        parent_node = self._get_node_parent(node)
        if parent_node is None:
            return parent_obj
        return self._get_abs_parent_obj(parent_node, parent_obj=parent_obj)

    def _get_node_parent(self, node):
        if 'parent_id' not in node:
            return None
        parent_node = self.id_map.get(node['parent_id'])
        if not parent_node:
            return None
        return parent_node

    # -----
    # Record API
    # -----

    def save_data(self, data):
        if len(self.activated_group) > 0:
            # TODO use self.save_to_group_id
            target_group_id = list(self.activated_group.keys())[0]
            self.add_data(target_group_id, data)
        else:
            self.add_data('tmp_group', data)

    # -----
    # Editor
    # -----

    def update_group(self, _id, data, save=True):
        ignore_keys = ['id', 'parent_id', 'type', 'children']
        node = self.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)

        update_data = {k: data[k] for k in data if k not in ignore_keys}
        node.update(update_data)

        delete_keys = [k for k in node if k not in data and k not in ignore_keys]
        for key in delete_keys:
            node.pop(key)

        # labels are ordered, sorted by label name
        if 'label' not in node:
            node['label'] = []
        elif 'label' in node and isinstance(node['label'], list):
            node['label'].sort(key=lambda x:x.get('name', '').lower())
        if save:
            self._adapter._update_group(node)

    def update_data(self, _id, data):
        node = self.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)
        node['name'] = data['name']
        self._adapter._update_data(data)
        self._adapter._update_group(node)

    # -----
    # Snapshot
    # -----

    def _write_prop_to_custom_path(self, outfile_path, node):
        self._adapter._write_prop_to_custom_path(outfile_path, node)

    def _write_file_to_custom_path(self, outfile_path, file_content):
        self._adapter._write_file_to_custom_path(outfile_path, file_content)

    def decompress_snapshot(self):
        return self._adapter.decompress_snapshot()

    def import_snapshot(self, parent_id, snapshot_name, path=None):
        return self._adapter.import_snapshot(parent_id, snapshot_name)

    def export_snapshot_from_event(self, event_json):
        return self._adapter.export_snapshot_from_event(event_json)

    def export_snapshot_from_dm(self, node_id):
        return self._adapter.export_snapshot_from_dm(node_id)

    def remove_tmp_snapshot_file(self, files):
        return self._adapter.remove_tmp_snapshot_file(files)

# -----------------
# Exceptions
# -----------------

class DumpPropError(Exception):
    pass



class RootNotSet(Exception):
    pass


class RootPathNotExists(Exception):
    pass


class RootPathIsNotDir(Exception):
    pass


class LyrebirdPropNotExists(Exception):
    pass


class DataNotFound(Exception):
    pass


class DataObjectCannotContainAnyOtherObject(Exception):
    pass


class DataObjectSouldBeADict(Exception):
    pass


class IDNotFound(Exception):
    pass


class NoneClipbord(Exception):
    pass


class NonePropFile(Exception):
    pass


class TooMuchPropFile(Exception):
    pass


class NodeExist(Exception):
    pass




from flask_restful import Resource
from lyrebird.mock import context
from flask import request, jsonify
from lyrebird import application


class Conf(Resource):
    """
    Lyrebird 及 插件 配置文件获取和修改
    """

    def get(self):
        return jsonify(application.config.raw())

    def put(self):
        try:
            context.application.conf = request.get_json()
            context.application.save()
            return context.make_ok_response()
        except Exception as e:
            return context.make_fail_response(str(e))



from lyrebird import application
from .. import checker


class CheckerEventHandler:

    def __call__(self, channel, *args, **kw):
        def func(origin_func):
            if not checker.scripts_tmp_storage.get(checker.TYPE_EVENT):
                checker.scripts_tmp_storage[checker.TYPE_EVENT] = []
            checker.scripts_tmp_storage[checker.TYPE_EVENT].append({
                'name': origin_func.__name__,
                'func': origin_func,
                'channel': channel
            })
            return origin_func
        return func

    def issue(self, title, message):
        notice = {
            "title": title,
            "message": message
        }
        application.server['event'].publish('notice', notice)

    def publish(self, channel, message, *args, **kwargs):
        application.server['event'].publish(channel, message, *args, **kwargs)

    @staticmethod
    def register(func_info):
        application.server['event'].subscribe(func_info['channel'], func_info['func'])

    @staticmethod
    def unregister(func_info):
        application.server['event'].unsubscribe(func_info['channel'], func_info['func'])

event = CheckerEventHandler()



from flask import send_file
from flask_restful import Resource, request
from lyrebird.mock import context

class SnapshotExport(Resource):
    def get(self, group_id):
        file = context.application.data_manager.export_snapshot_from_dm(group_id)
        return send_file(file)

    def post(self):
        file = context.application.data_manager.export_snapshot_from_event(request.json)
        return send_file(file)



from pathlib import Path
from os import path
import codecs
import json
from packaging import version
import jinja2
from lyrebird import log as nlog


logger = nlog.get_logger()


config_template = {
    "version": "0.10.5",
    "proxy.filters": [],
    "proxy.port": 4272,
    "mock.port": 9090,
    "mock.data": path.join("{{current_dir}}", "mock_data", "personal"),
    "mock.proxy_headers": {
        "scheme": "MKScheme",
        "host": "MKOriginHost",
        "port": "MKOriginPort"
    }
}


class ConfigManager():
    ROOT = Path('~/.lyrebird').expanduser()
    DEFAULT_FILENAME = 'conf.json'
    BASE_CONFIG = ROOT/DEFAULT_FILENAME

    def __init__(self, conf_path=None):
        self.config = config_template
        self.config_root = self.ROOT
        self.conf_file = self.BASE_CONFIG

        self.update_base_config()
        self.read_config()
        if conf_path:
            self.update_conf_source(conf_path)

    def update_conf_source(self, path):
        input_path: Path = Path(path).expanduser().absolute()
        if input_path.is_dir():
            input_root = input_path
            input_file = input_path / self.DEFAULT_FILENAME
        else:
            input_root = input_path.parent
            input_file = input_path

        if not input_file.exists():
            logger.error(f'Config {input_file} not found!')
        else:
            self.config_root = input_root
            self.conf_file = input_file
            self.read_config()

    def read_config(self):
        template_env = jinja2.Environment(loader=jinja2.FileSystemLoader(str(self.config_root)))
        template = template_env.get_template(self.conf_file.name)
        current_dir = str(self.config_root)
        download_dir = str(self.ROOT/'downloads')
        conf_str = template.render(current_dir=json.dumps(current_dir).strip('"'), download_dir=json.dumps(download_dir).strip('"'))
        loaded_config = json.loads(conf_str)
        self.config.update(loaded_config)

    def write_config(self):
        self.config_root.mkdir(parents=True, exist_ok=True)
        with codecs.open(self.conf_file, 'w', 'utf-8') as f:
            f.write(json.dumps(self.config, indent=4, ensure_ascii=False))

    def update_base_config(self):
        if self.BASE_CONFIG.exists() and self.BASE_CONFIG.is_file():
            with codecs.open(self.BASE_CONFIG, 'r', 'utf-8') as f:
                base_conf = json.load(f)
                if version.parse(base_conf.get('version', '0.0.0')) < version.parse(config_template.get('version', '0.0.0')):
                    self.write_config()
        else:
            self.write_config()


class ConfigException(Exception):
    pass



from pathlib import Path
import jinja2
import codecs

here = Path(__file__).parent
demo_project = Path(here) / 'lyrebird_plugin_demo'


def make_plugin_project(project_name, project_dir):
    # Context
    ctx = {
        'project_name': project_name
    }
    # Make project dir
    Path(project_dir).expanduser().absolute().mkdir(parents=True, exist_ok=True)
    # Copy files to target dir
    _cp_dir(ctx, demo_project, project_dir)


def _cp_dir(ctx, _path, target):
    target_dir = Path(target)
    if target_dir.name == '$_plugin_name':
        target = target_dir.parent/ctx['project_name']
    # Make target directory
    Path(target).mkdir(parents=True, exist_ok=True)
    print(f'MKDIR {target}')
    # cp subfile to target dir
    for subfile in Path(_path).iterdir():
        if subfile.is_dir():
            _cp_dir(ctx, subfile, Path(target)/subfile.name)
        elif subfile.is_file():
            _cp_file(ctx, subfile, Path(target)/subfile.name)


def _cp_file(ctx, _path, target):
    print(f'CP file {_path} to {target}')
    if Path(target).suffix in ['.py', '.txt', '.md', '.in', '.js']:
        with codecs.open(_path, 'r', 'utf-8') as f_in, codecs.open(target, 'w', 'utf-8') as f_out:
            in_file_string = f_in.read()
            file_template = jinja2.Template(in_file_string)
            in_file_string = file_template.render(project_name=ctx['project_name'])
            f_out.write(in_file_string)
    else:
        with codecs.open(_path, 'rb') as f_in, codecs.open(target, 'wb') as f_out:
            f_out.write(f_in.read())



"""
Script for mitmdump

Redirect request from proxy server to mock server
"""

from mitmproxy import http
from lyrebird import log
import os
import json
import logging
import re

_logger = log.get_logger()
_logger.setLevel(logging.INFO)

PROXY_PORT = int(os.environ.get('PROXY_PORT'))
PROXY_FILTERS = json.loads(os.environ.get('PROXY_FILTERS'))


def to_mock_server(flow: http.HTTPFlow):
    # mock path 为/mock开头加上原始url
    flow.request.path = '/mock/' + flow.request.url
    # mock scheme 统一为http
    flow.request.scheme = 'http'
    # mock server port
    flow.request.port = PROXY_PORT
    # mock server ip
    flow.request.host = '127.0.0.1'
    # device real ip
    address = flow.client_conn.address[0]
    # 获取的address是IPv6（内嵌IPv4地址表示法），需要获取IPv4地址，需要做以下处理
    if address.startswith('::ffff:'):
        address = address.split('::ffff:')[1]
    
    flow.request.headers['Lyrebird-Client-Address'] = address
    flow.request.headers['Proxy-Raw-Headers'] = json.dumps({name: flow.request.headers[name] for name in flow.request.headers}, ensure_ascii=False)

    _logger.info('Redirect-> %s' % flow.request.url[:100])


def request(flow: http.HTTPFlow):
    _logger.info(flow.request.url[:100])
    if 'mitm.it' in flow.request.url:
        # Support mitm.it
        return

    if not PROXY_FILTERS:
        to_mock_server(flow)
        return

    for _filter in PROXY_FILTERS:
        if re.search(_filter, flow.request.url):
            to_mock_server(flow)
            break


def responseheaders(flow):
    """
    Enables streaming for all responses.
    This is equivalent to passing `--set stream_large_bodies=1` to mitmproxy.
    """
    if 'mitm.it' in flow.request.url:
        # Support mitm.it
        flow.response.stream = False
        return
    flow.response.stream = True



import io
import qrcode
import binascii
from jinja2 import Template
from lyrebird.application import config

def _format_parameters(link):
    params = {
        'ip': config.get('ip'),
        'port': config.get('mock.port')
    }
    link_template = Template(link)
    formated_link = link_template.render(params)
    return formated_link

def get_qrcode_img(link):
    formated_link = _format_parameters(link)

    img = qrcode.make(formated_link, border=1)
    img_buffer = io.BytesIO()
    img.save(img_buffer, 'png')
    img_data_bytes = img_buffer.getvalue()
    img_buffer.close()

    img_data = (b"data:image/png;base64," + binascii.b2a_base64(img_data_bytes)).decode()
    return img_data



from flask import jsonify

"""
Lyrebird context

"""


def make_ok_response(**kwargs):
    ok_resp = {
        "code": 1000,
        "message": "success"
    }
    ok_resp.update(kwargs)
    return jsonify(ok_resp)


def make_fail_response(msg, **kwargs):
    fail_resp = {
        "code": 3000,
        "message": msg
    }
    fail_resp.update(kwargs)
    return jsonify(fail_resp)


# Lyrebird status contains: 'READY' and 'INITING'
status = 'INITING'

_cm = None
_src = None


server = {}
plugins = {}
active_menu = {}

snapshot_import_uri = ''


notice = None
checkers = {}

on_request = []
on_response = []
on_request_upstream = []
on_response_upstream = []

encoder = []
decoder = []

labels = None

encoders_decoders = None

def start_server():
    for name in server:
        server[name].start()


def stop_server():
    for name in server:
        server[name].stop()


class ConfigProxy:

    def get(self, k, default=None):
        return _cm.config.get(k, default)

    def __setitem__(self, k, v):
        _cm.config[k] = v

    def __getitem__(self, k):
        return _cm.config[k]

    def raw(self):
        return _cm.config


config = ConfigProxy()

# statistics reporter
reporter = None


def root_dir():
    if _cm:
        return _cm.ROOT



import math
import time


def convert_size(size_bytes):
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])


def convert_time(duration):
    if duration < 1:
        return str(round(duration * 1000)) + 'ms'
    else:
        return str(round(duration, 2)) + 's'


def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        print(f'{method.__name__} execution time {(te-ts)*1000}')
        return result
    return timed



import os
import uuid
import time
import codecs
import tarfile
import requests
from lyrebird import log
from lyrebird import application
from urllib.parse import urlparse
from pathlib import Path
logger = log.get_logger()


class SnapshotHelper():

    def get_snapshot_path(self):
        snapshot_repositories = application._cm.ROOT / "snapshot"
        if not snapshot_repositories.exists():
            snapshot_repositories.mkdir()
        temp_dir_name = f"{time.strftime('%Y-%m-%d-%H-%M-%S', time.localtime())}-{str(uuid.uuid4())}"
        snapshot_path = snapshot_repositories / temp_dir_name
        snapshot_path.mkdir()
        return snapshot_path

    def save_compressed_file(self, snapshot_path):
        snapshot_import_uri = application.snapshot_import_uri
        if not snapshot_import_uri:
            raise SnapshotImportUriNotFound
        parsed_uri = urlparse(snapshot_import_uri)
        if "http" in parsed_uri.scheme:
            resp = requests.get(snapshot_import_uri)
            with open(f"{snapshot_path}.lb", "wb") as f:
                for chunck in resp.iter_content():
                    f.write(chunck)
        elif "file" in parsed_uri.scheme:
            with codecs.open(parsed_uri.path, "rb") as inputfile, codecs.open(f"{snapshot_path}.lb", "wb") as outputfile:
                read_content = inputfile.read()
                outputfile.write(read_content)
        else:
            raise SnapshotImportUriSchemeIllegal

    def decompress_snapshot(self, compress_dir, decompress_dir):
        tf = tarfile.open(compress_dir)
        tf.extractall(decompress_dir)
        tf.close()
        logger.debug(f"decompress [{compress_dir}] to [{decompress_dir}] success")

    def compress_snapshot(self, output_path, dir_in):
        cur_path = Path.cwd()
        full_path_in = cur_path.joinpath(dir_in)
        os.chdir(full_path_in)
        tar = tarfile.open(f"{output_path}.lb", "w:gz")
        for root, dirs, files in os.walk(full_path_in):
            logger.debug(f"compress use root: {root}, dirs: {dirs}, files: {files}")
            for file in files:
                tar.add(file, recursive=False)
        tar.close()
        os.chdir(cur_path)
        logger.debug(f"compress snapshot : ['{output_path}.lb'] success")

    def get_data_id_map(self, node, data_id_map):
        if 'id' in node and node['type'] == 'data':
            data_id_map[node['id']] = node
        if 'children' in node and node['type'] == 'group':
            for child in node['children']:
                self.get_data_id_map(child, data_id_map)


# -----------------
# Exceptions
# -----------------
class SnapshotImportUriNotFound(Exception):
    pass


class SnapshotImportUriSchemeIllegal(Exception):
    pass



from flask_restful import Resource
from lyrebird.mock import context
from lyrebird import application
from urllib.parse import urlencode, unquote
from flask import request, Response
import json

class Flow(Resource):
    """
    当前请求单条数据
    """

    def get(self, id):
        for item in context.application.cache.items():
            if item['id'] == id:
                # Import decoder for decoding the requested content
                display_item = {}
                application.encoders_decoders.decoder_handler(item, output=display_item)
                return application.make_ok_response(data=display_item)
        return application.make_fail_response(f'Request {id} not found')


class FlowList(Resource):
    """
    当前请求列表
    """

    def get(self):
        ignore_host = context.application.selected_filter.get('ignore', []) if context.application.selected_filter else []
        all_items = context.application.cache.items()[::-1]
        req_list = []
        for item in all_items:
            if item['request'].get('host') in ignore_host:
                continue
            info = dict(
                id=item['id'],
                size=item['size'],
                duration=item['duration'],
                start_time=item['start_time'],
                request=dict(
                    url=item['request'].get('url'),
                    scheme=item['request'].get('scheme'),
                    host=item['request'].get('host'),
                    path=item['request'].get('path'),
                    params=unquote(urlencode(item['request']['query'])),
                    method=item['request'].get('method')
                ),
                response=dict(
                    code=item['response']['code'],
                    mock=item['response']['headers'].get('lyrebird', 'proxy'),
                    modified=item['request']['headers'].get('lyrebird_modified') or item['response']['headers'].get('lyrebird_modified', '')
                )if item.get('response') else {},
                action=item.get('action', [])
            )
            # Add key `proxy_response` into info only if item contains proxy_response
            if item.get('proxy_response'):
                info['proxy_response'] = {
                    'code': item['proxy_response']['code']
                }
            req_list.append(info)

        return Response(json.dumps(req_list, ensure_ascii=False), mimetype='application/json', status=200)

    def delete(self):
        _ids = request.json.get('ids')
        if _ids:
            context.application.cache.delete_by_ids(*_ids)
        else:
            context.application.cache.clear()
        context.application.socket_io.emit('action', 'delete flow log')
        return application.make_ok_response()

    def post(self):
        _ids = request.json.get('ids')
        record_items = []
        for _id in _ids:
            for item in context.application.cache.items():
                if _id == item['id']:
                    record_items.append(item)
                    break
        dm = context.application.data_manager

        for flow in record_items:
            dm.save_data(flow)

        return application.make_ok_response()


class FlowFilter(Resource):

    def get(self):
        filters = context.application.filters
        selected_filter = context.application.selected_filter
        return application.make_ok_response(selected_filter=selected_filter, filters=filters)

    def put(self):
        selected_filter_name = request.json.get('name')
        if not selected_filter_name:
            context.application.selected_filter = None
            return application.make_ok_response()

        for f in context.application.filters:
            if f['name'] == selected_filter_name:
                context.application.selected_filter = f
                return application.make_ok_response()

        return application.make_fail_response(f'Flow filter {selected_filter_name} not found!')



import argparse
import webbrowser
import json
import socket
import threading
import signal
import os
import platform
import traceback
from pathlib import Path
from lyrebird import log
from lyrebird import application
from lyrebird.config import ConfigManager
from lyrebird.mock.mock_server import LyrebirdMockServer
from lyrebird.proxy.proxy_server import LyrebirdProxyServer
from lyrebird.event import EventServer
from lyrebird.task import BackgroundTaskServer
from lyrebird.notice_center import NoticeCenter
from lyrebird.mock.dm.label import LabelHandler
from lyrebird.mock.handlers.encoder_decoder_handler import EncoderDecoder
from lyrebird.db.database_server import LyrebirdDatabaseServer
from lyrebird.plugins import PluginManager
from lyrebird.checker import LyrebirdCheckerServer
from lyrebird import version
from lyrebird import reporter
from lyrebird import project_builder
from packaging.version import parse as vparse


logger = log.get_logger()


def main():
    """
    Command line main entry

    Start lyrebird

    * start in default config
    ```
    lyrebird
    ```
    * start with verbose mode
    ```
    lyrebird -v
    ```
    * start without open a web browser
    ```
    lyrebird -b
    ```
    * start with a specified config file
    ```
    lyrebird -c /path/to/your/config/file
    ```
    * start with multipart args
    ```
    lyrebird -v --mock 8080 -c /path/to/your/config/file
    ```
    """
    parser = argparse.ArgumentParser(prog='lyrebird')

    parser.add_argument('-V', '--version', dest='version', action='store_true', help='show lyrebird version')
    parser.add_argument('-v', dest='verbose', action='count', default=0, help='Show verbose log')
    parser.add_argument('--ip', dest='ip', help='Set device ip')
    parser.add_argument('--mock', dest='mock', type=int, help='Set mock server port, default port is 9090')
    parser.add_argument('--proxy', dest='proxy', type=int, help='Set proxy server port, default port is 4272')
    parser.add_argument('--data', dest='data', help='Set data dir, default is "./data/"')
    parser.add_argument('-b', '--no_browser', dest='no_browser',
                        action='store_true', help='Start without open a browser')
    parser.add_argument('-c', '--config', dest='config',
                        help='Start with a config file. Default is "~/.lyrebird/conf.json"')
    parser.add_argument('--log', dest='log', help='Set output log file path')
    parser.add_argument('--script', action='append', help='Set a checker script path')
    parser.add_argument('--plugin', action='append', help='Set a plugin project path')
    parser.add_argument('--database', dest='database', help='Set a database path. Default is "~/.lyrebird/lyrebird.db"')

    subparser = parser.add_subparsers(dest='sub_command')

    gen_parser = subparser.add_parser('gen')
    gen_parser.add_argument('path', help='Create plugin project')

    args = parser.parse_args()

    if args.version:
        print(version.LYREBIRD)
        return

    Path('~/.lyrebird').expanduser().mkdir(parents=True, exist_ok=True)

    if args.config:
        application._cm = ConfigManager(conf_path=args.config)
    else:
        application._cm = ConfigManager()

    # set current ip to config
    if args.ip:
        application._cm.config['ip'] = args.ip
    else:
        try:
            application._cm.config['ip'] = _get_ip()
        except socket.gaierror as e:
            logger.error(f'Failed to get local IP address, error occurs on {e}')

    # init file logger after config init
    application._cm.config['verbose'] = args.verbose
    log.init(args.log)

    if args.mock:
        application._cm.config['mock.port'] = args.mock
    if args.proxy:
        application._cm.config['proxy.port'] = args.proxy
    if args.data:
        application._cm.config['mock.data'] = str(Path(args.data).expanduser().absolute())

    logger.debug(f'Read args: {args}')

    if args.sub_command == 'gen':
        logger.debug('EXEC: Plugin project generator')
        gen(args)
    else:
        logger.debug('EXEC: LYREBIRD START')
        run(args)


def run(args: argparse.Namespace):
    sys_name = platform.system()
    if sys_name.lower() != 'windows':
        import resource
        # Set file descriptors
        try:
            resource.setrlimit(resource.RLIMIT_NOFILE, (8192, 8192))
        except Exception:
            traceback.print_exc()
            logger.warning('Set file descriptors failed\nPlease set it by your self, use "ulimit -n 8192" with root account')

    # show current config contents
    print_lyrebird_info()
    config_str = json.dumps(application._cm.config, ensure_ascii=False, indent=4)
    logger.warning(f'Lyrebird start with config:\n{config_str}')

    # Main server
    application.server['event'] = EventServer()

    application.server['task'] = BackgroundTaskServer()
    application.server['proxy'] = LyrebirdProxyServer()
    application.server['mock'] = LyrebirdMockServer()
    application.server['db'] = LyrebirdDatabaseServer(path=args.database)
    application.server['plugin'] = PluginManager()
    application.server['checker'] = LyrebirdCheckerServer()

    application.start_server()

    # int statistics reporter
    application.reporter = reporter.Reporter()
    reporter.start()
    # activate notice center
    application.notice = NoticeCenter()

    # init label handler
    application.labels = LabelHandler()

    # init encoder&decoder
    application.encoders_decoders = EncoderDecoder()

    # load debug plugin
    # TODO
    plugin_manager = application.server['plugin']
    if args.plugin:
        plugin_manager.plugin_path_list += args.plugin
    plugin_manager.reload()

    # load debug script
    if args.script:
        application.server['checker'].load_scripts(args.script)

    # auto open web browser
    if not args.no_browser:
        webbrowser.open(f'http://localhost:{application.config["mock.port"]}')

    # Lyrebird status contains: 'READY' and 'INITING'
    application.status = 'READY'

    # stop event handler
    def signal_handler(signum, frame):
        reporter.stop()
        application.stop_server()
        threading.Event().set()
        logger.warning('!!!Ctrl-C pressed. Lyrebird stop!!!')
        os._exit(1)

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    threading.Event().wait()


def gen(args):
    parent_path = args.path
    project_name = input('Please input your project name:')
    if not project_name:
        print('Not set project name')
        return
    project_builder.make_plugin_project(project_name, parent_path+'/'+project_name)


def _get_ip():
    """
    Get local ip from socket connection

    :return: IP Addr string
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('bing.com', 80))
    return s.getsockname()[0]


def print_lyrebird_info():
    logo = [
        "",
        "",
        "     _                    _     _         _ ",
        "    | |                  | |   (_)       | |",
        "    | |    _   _ _ __ ___| |__  _ _ __ __| |",
        "    | |   | | | | '__/ _ \\ '_ \\| | '__/ _' |",
        "    | |___| |_| | | |  __/ |_) | | | | (_| |",
        "    \\_____/\\__, |_|  \\___|_.__/|_|_|  \\__,_|",
        "            __/ |                           ",
        "           |___/                            ",
        "",
        f"                   v{version.VERSION}",
        "",
        "",
        ""
    ]
    logo_str = '\n'.join(logo)
    # Custom log level 60  : NOTICE
    logger.log(60, logo_str)



import hashlib
from lyrebird.mock import context


class LabelHandler:

    def __init__(self):
        self.label_map = {}
        self.isolated_label_ids = set()
        self.default_color = '#808695'
        self.default_description = ''

    def get_label(self, node):
        # collect isolated label
        self.isolated_label_ids = set()
        isolated_labels = {}
        for label_id, label in self.label_map.items():
            if not label['groups']:
                self.isolated_label_ids.add(label_id)
                isolated_labels[label_id] = label

        # init label_map
        self.label_map = {}
        self._collect_label(node)

        # add isolated_label
        for key in self.label_map:
            if key in isolated_labels:
                isolated_labels.pop(key)
        self.label_map.update(isolated_labels)

    def _make_label_info(self, group):
        group_id = group['id']
        label_list = group.get('label')
        if not isinstance(label_list, list):
            return

        for label in label_list:
            name = label.get('name')
            if not name:
                continue
            label_id = self._get_label_name_md5(label)
            if self.label_map.get(label_id):
                self.label_map[label_id]['groups'].append(group_id)
            else:
                self.label_map[label_id] = {
                    'id': label_id,
                    'name': label['name'],
                    'color': label.get('color') or self.default_color,
                    'description': label.get('description') or '',
                    'groups': [group_id]
                }

    def _collect_label(self, node):
        if node['type'] == 'group' and 'label' in node:
            self._make_label_info(node)
        if 'children' in node:
            for child in node['children']:
                self._collect_label(child)

    def create_label(self, label):
        label_name = label.get('name')
        if not label_name:
            raise LabelNameNotFound

        label_id = self._get_label_name_md5(label)
        new_label = {
            label_id: {
                'id': label_id,
                'name': label['name'],
                'color': label.get('color') or self.default_color,
                'description': label.get('description') or '',
                'groups': []
            }
        }
        self.isolated_label_ids.add(label_id)
        self.label_map.update(new_label)

    def update_label(self, update_label):
        target_label_id = update_label['id']
        update_name = update_label.get('name')
        update_color = update_label.get('color') or self.default_color
        update_description = update_label.get('description') or self.default_description
        if not update_name:
            raise LabelNameNotFound

        target_label = self.label_map.get(target_label_id)
        groups = target_label.get('groups') or []
        for group_id in groups:
            _group = context.application.data_manager.id_map.get(group_id)
            label_list = _group.get('label')
            if not isinstance(label_list, list):
                raise LabelTypeIsNotList(f'Unexpected label value type: {type(label_list)}')

            for label in label_list:
                label_id = self._get_label_name_md5(label)
                if label_id != target_label_id:
                    continue
                label['name'] = update_name
                label['color'] = update_color
                label['description'] = update_description
                break

            context.application.data_manager.update_group(group_id, _group)

        if groups:
            context.application.data_manager.reload()

        # update label_map
        self.label_map[target_label_id]['name'] = update_name
        self.label_map[target_label_id]['color'] = update_color
        self.label_map[target_label_id]['description'] = update_description

        # update label_map id if label_id is changed
        new_label_id = self._get_label_name_md5(update_label)
        if new_label_id != target_label_id:
            self.label_map[target_label_id]['id'] = new_label_id
            self.label_map[new_label_id] = self.label_map[target_label_id]
            self.label_map.pop(target_label_id)

    def delete_label(self, target_label_id):
        target_label = self.label_map.get(target_label_id)
        groups = target_label.get('groups') or []
        for group_id in groups:
            _group = context.application.data_manager.id_map.get(group_id)
            label_list = _group.get('label')
            if not isinstance(label_list, list):
                raise LabelTypeIsNotList(f'Unexpected label value type: {type(label_list)}')

            for label in label_list[::-1]:
                label_id = self._get_label_name_md5(label)
                if label_id != target_label_id:
                    continue
                label_list.remove(label)
                break

            context.application.data_manager.update_group(group_id, _group)

        if groups:
            context.application.data_manager.reload()

        # update data_map
        self.label_map.pop(target_label_id)
        if target_label_id in self.isolated_label_ids:
            self.isolated_label_ids.remove(target_label_id)

    def _get_label_name_md5(self, label):
        keyword = label['name']
        md5_module = hashlib.sha224()
        md5_module.update(bytes(keyword, encoding="utf-8"))
        return md5_module.hexdigest()


class LabelNotFound(Exception):
    pass


class LabelTypeIsNotList(Exception):
    pass


class LabelNameNotFound(Exception):
    pass
